import random
import orjson
import typer
import rich
import sys
import os
import re

from typing import List, Tuple, Dict, Union, Optional, Any, Iterator, TypeVar
from iotree.core.render.tables import tableFromRecords
from iotree.core.render.funcs import basic_pbar
from pathlib import Path

from .utils.render import console, errors
from .core.isql_config import init_isql
from .core.isql import iSQL

from pisql.utils.paths import (
    base_dir, tests_dir, package_dir,
    assets_dir, data_dir
)

from .cli.execute import exe
from .cli.run import run
from .cli.config import config, iSqlCFG

isql = init_isql()

app = typer.Typer(
    name="pisql",
    help="A simple but rich command line interface for Sybase ASE",
    no_args_is_help=True,
    rich_help_panel="rich",
    rich_markup_mode="rich"
)

dev = typer.Typer(
    name="dev",
    help=f"""Developer tools 💻.
    [yellow][bold]Aliases:[/] d, dev[/]
    """,
    rich_markup_mode="rich",
    rich_help_panel="rich",
    no_args_is_help=True,
)


app.add_typer(exe, name="execute")
app.add_typer(exe, name="exec", hidden=True)
app.add_typer(exe, name="e", hidden=True)
app.add_typer(exe, name="x", hidden=True)

app.add_typer(dev, name="dev", no_args_is_help=True)
app.add_typer(dev, name="d", no_args_is_help=True, hidden=True)

app.add_typer(run, name="run", no_args_is_help=True, chain=True, help="Run one or many  queries in a row with a rich progress bar  " )
app.add_typer(run, name="q", no_args_is_help=True, hidden=True, chain=True, )
app.add_typer(run, name="::", no_args_is_help=True, hidden=True, chain=True)

app.add_typer(config, name="config", no_args_is_help=True)
app.add_typer(config, name="conf", no_args_is_help=True, hidden=True)
app.add_typer(config, name="cf", no_args_is_help=True, hidden=True)

@dev.command(name="test", help="Run tests 🧪")
def test(
    test_type: str = typer.Argument(..., help="The type of test to run 🧪"),
    index: int = typer.Option(None, "-i", "-idx", "--index", help="The index of the subtest to run [dim](if your test has many options)[/] 🧪"),
    name: str = typer.Option(None, "-n", "--name", help="The name (a part of it) of the subtest to run [dim](if your test has many options)[/] 🧪"),
    ) -> None:
    """Run tests 🧪"""
    test_type = test_type.lower()
    if index is not None and name is not None:
        errors.print("[bold red]You can't specify both index and name[/]", style="bold red")
        errors.print("[dim yellow]Using `name`.[/]")
        index = None
        selector = name
    elif index is not None:
        selector = index
    elif name is not None:
        selector = name
    else:
        selector = None
    
    if test_type in ["parse", "parsing", "text"]:
        raw_outs = {
            name: open( data_dir / name, "r").read() for name in os.listdir(data_dir)
        }
        if selector is None:
            selector = random.randint(0, len(raw_outs) - 1)

        target = [
            name for name in raw_outs.keys() if selector in name
        ]

        if not len(target):
            raise typer.BadParameter(f"Couldn't find a test with name {selector}")

@dev.command(name="view", help="See the raw data 📰 🧪")
def test(
    name: str = typer.Argument(..., help="The name (a part of it) of the subtest to run [dim](if your test has many options)[/] 🧪"),
    head: int = typer.Option(500, "-h", "--head", help="The number of lines to show [dim](if your file is huge !)[/] 🧪"),
    raw: bool = typer.Option(False, "-ra", "--raw", help="Whether to show the raw data 📰 🧪"),
    remove_whitespaces: bool = typer.Option(False, "-rw", "--rm-ws", "--remove-whitespaces", help="Remove whitespaces from the output [dim](if your data uses them too much !)[/] 🧪"),
    columns: bool = typer.Option(False, "-c", "--cols", "--columns", help="Whether to only display the parsed columns 🏦"),
    ) -> None:

    raw_outs = {
            name: open( data_dir / name, "r").read() for name in os.listdir(data_dir)
        }
    target = [
        key for key in raw_outs.keys() if name in key
    ]

    if not len(target):
        raise typer.BadParameter(f"Couldn't find a raw datafile with name {name}")
    
    target = target.pop()
    raw_out = raw_outs[target]

    if raw:
        idx = 0
        for i in range(head):
            idx = raw_out.find("'", idx + 1)
        console.print(raw_out[:idx])
        sys.exit(0)

    if columns:
        df = isql.strRowColSplitter(raw_out)
        console.print(df.columns)
        sys.exit(0)
    else:
        df = isql.strRowColSplitter(raw_out)
        records = orjson.loads(df.write_json(row_oriented=True))
        console.print(tableFromRecords(records[:head], theme="default"))

    if remove_whitespaces:
        raw_out = re.sub(r"([ \t\n]+)", " ", raw_out)
    
    sys.exit(0)



@dev.command(name="list", help="List the test queries 📰🗄️")
def list():
    """List the test queries 📰🗄️"""
    queries_dir = assets_dir / "sql"
    console.print(f"[bold magenta]Test queries:[/]")
    
    for file in os.listdir(queries_dir):
        console.print(f"[yellow] \u2736 📰 {file}[/]")
    
    console.print(f"[bold magenta]Total:[/] {len(os.listdir(queries_dir))} 📰🗄️")
    console.print(f"Let's go to work ! 🚀🚀🚀")

if __name__ == "__main__":
    typer.run(app)


import sys
from rich.prompt import Confirm, Prompt
from ..core.isql import iSQL
from ..utils.render import console, errors
from ..utils.local_store import iSqlConfig

iSqlCFG = iSqlConfig('pisql')

def interactive_init():
    config_store = iSqlCFG.getCU()
    yes = Confirm.ask("Do you want to configure a server now?", default=True)
    if yes:
        sql_server = Prompt.ask("Server name")
        sql_user = Prompt.ask("Username")
        sql_password = Prompt.ask("Password", password=True)
        sql_db = Prompt.ask("Database")
        config_store["server"] = sql_server
        config_store["servers"] = {
            sql_server: {
                "server": sql_server,
                "user": sql_user,
                "password": sql_password,
                "db": sql_db
            }
        }
        iSqlCFG.setCU(config_store)
    else:
        console.print(
            f"""[bold red]Exiting...[/]
            [dim yellow]You can configure a server later with [/][bold]pisql config[/]
            """
            )
        sys.exit(0)

def init_isql() -> iSQL:

    incomplete = False
    inexistant = False
    
    config_store = iSqlCFG.getCU()
    
    if not "server" in config_store or not "servers" in config_store:
        inexistant = True
        incomplete = True
    else:
        if not "server" in config_store:
            errors.print(
                f"""[bold red]No default server found in config[/]
                [dim yellow]Please specify a server with [/][bold]pisql config[/]
                """
                )
            incomplete = True
        elif not config_store["server"] in config_store["servers"]:
            errors.print(
                f"""[bold red]Default server not found in config[/]
                [dim yellow]Please specify a server with [/][bold]pisql config[/]
                """
                )
            incomplete = True
        else:

            config = config_store["servers"][config_store["server"]]
            if any(
                not x in config
                for x in ["server", "user", "password", "db"]
            ):
                errors.print(
                    f"""[bold red]Incomplete config[/]
                    [dim yellow]Please specify a server with [/][bold]pisql config[/]
                    """
                    )
                incomplete = True
            else:
                return iSQL(
                    server=config["server"],
                    user=config["user"],
                    password=config["password"],
                    db=config["db"]
                )
    if inexistant:
        errors.print(
            f"[bold]Welcome to [green]pisql[/][/]\n" +
            f"[bold][magenta]pisql[/bold] is a simple but rich command line interface for Sybase ASE[/]\n" +
            f"❗❗[blink bold red]No server found in config[/]\n[bold red]> Please configure a server first[/]"
            )
    if incomplete:
        console.print("[bold red]Incomplete config.[/] We will now ask you to configure a server")
        interactive_init()
        return init_isql()



import polars as pl
import typer

from pathlib import Path
from typing import List, Tuple, Dict, Union, Optional, Any, Iterator, TypeVar
from pisql.core.isql import iSQL
from pisql.core.isql_config import init_isql
from pisql.utils.paths import (
    base_dir, tests_dir, package_dir
)

isql = init_isql()

exe = typer.Typer(
    name="execute",
    help=f"""Execute a query against a Sybase ASE database . Accepts a .sql file ",
    [yellow][bold]Alias:[/] exec, e, x[/]  
    [green][bold]Example:[/]  
    >>> pisql x name.sql[/]
    """,
    rich_markup_mode="rich",
    rich_help_panel="rich",
)


@exe.callback(invoke_without_command=True, no_args_is_help=True, help="Execute a query against a Sybase ASE database . Accepts a .sql file ")
def executer(
    file: str = typer.Argument(..., help="The .sql file  to execute"),
    out: str = typer.Option(None, "-o", "--out", "--out-file", help="The output file , defaults to rich table  on console"),
    override_default_dir: str = typer.Option(None, "-od", "--override-default-dir", help="Override the default directory . Defaults to `~/pisql`"),
    ) -> None:
    """Execute a query  against a Sybase ASE database . Accepts a .sql file \n[yellow][bold]Aliases:[/] exec, e, x[/]
    
    Args:
        file (str): The .sql file  to execute
        out (str, optional): The output file , defaults to rich table  on console. Defaults to None.
        override_default_dir (str, optional): Override the default directory . Defaults to `~/pisql`. Defaults to None.

        Returns:
            None: None"""
    

    file = Path(file)

    if not override_default_dir:
        dir = Path(os.path.expanduser("~")) / "pisql" / "results"
        dir.mkdir(parents=True, exist_ok=True)
    else:
        dir = Path(override_default_dir) if override_default_dir != "." else Path.cwd()
        dir.mkdir(parents=True, exist_ok=True)

    with basic_pbar() as pbar:
        task = pbar.add_task(f"Executing {file.name}", total=None)
        df = isql.run_sql_file(file)

        pbar.update(task, advance=50)
        pbar.update(task, description=f"Writing results to {dir}")

        csv_file = dir / f"{file.stem}.csv"
        parquet_file = dir / f"{file.stem}.parquet"

        if out is None:
            csv_out = df.write_csv(file=csv_file)
            parquet_out = df.write_parquet(file=parquet_file,  compression="lz4")
            records = orjson.loads(df.write_json(row_oriented=True))
            console.print(tableFromRecords(records, theme="default"))
        else:
            outfile = dir / file.name
            csv_out = df.write_csv(file=csv_file)
            parquet_out = df.write_parquet(file=parquet_file, compression="lz4")
            if out.lower() in ["json", "excel"]:
                errors.print(f"[bold red]Output format {out} not supported yet with the base version.[/]", style="bold red")
                errors.print(f"[dim yellow]Please install `pandas` to get excel support[/]")
                sys.exit(1)
        pbar.update(task, advance=50)
        pbar.update(task, description=f"✅ Done writing results to {dir}", completed=True)
        sys.exit(0)



import os
from pathlib import Path

package_dir = Path(__file__).parent.parent
base_dir = package_dir.parent
tests_dir = base_dir / "tests"
assets_dir = package_dir / "assets"
data_dir = assets_dir / "data"

pisql_home = Path(os.path.expanduser("~")) / "pisql"